/*----------------------------------------------------------------------------------------------------
	AMD_HD3D.cpp

	AMD Quad Buffer Stereo Class for High Definition 3D (HD3D)

Copyright  2011 ADVANCED MICRO DEVICES, INC.  All Rights Reserved.

AMD is granting you permission to use this software and documentation (if any) (collectively, the Materials) pursuant to the terms and conditions of the Software Development Kit License Agreement included with the Materials.  This header does NOT give you permission to use the Materials or any rights under AMDs intellectual property.  Your use of any portion of these Materials shall constitute your acceptance of the terms and conditions of such Software Development Kit License Agreement.  If you do not agree to the terms and conditions of the Software Development Kit License Agreement, you do not have permission to use any portion of these Materials.  If you do not have a copy of the Software Development Kit License Agreement, contact your AMD representative for a copy.
CONFIDENTIALITY:  The Materials and all other related information, identified as confidential and provided to you by AMD shall be kept confidential in accordance with the terms and conditions of the Software Development Kit License Agreement.  If no Software Development Kit License Agreement exists between you and AMD, you agree to keep these Materials confidential and to not disclose it to any third party.
LIMITATION OF LIABILITY: THE MATERIALS ARE PROVIDED AS IS WITHOUT ANY EXPRESS OR IMPLIED WARRANTY OF ANY KIND INCLUDING WARRANTIES OF MERCHANTABILITY, NONINFRINGEMENT OF THIRD-PARTY INTELLECTUAL PROPERTY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE, OR THOSE ARISING FROM CUSTOM OF TRADE OR COURSE OF USAGE.  
FOR CLARIFICATION, THE ENTIRE RISK ARISING OUT OF USE OR PERFORMANCE OF THE MATERIALS REMAINS WITH YOU.  AMD DOES NOT WARRANT, GUARANTEE, OR MAKE ANY REPRESENTATIONS AS TO THE CORRECTNESS, ACCURACY, COMPLETENESS, QUALITY, OR RELIABILITY OF THE MATERIALS.  AMD DOES NOT WARRANT THAT OPERATION OF THE MATERIALS WILL BE UNINTERRUPTED OR ERROR-FREE.  YOU ARE RESPONSIBLE FOR DETERMINING THE APPROPRIATENESS OF USING THE SOFTWARE AND ASSUME ALL RISKS ASSOCIATED WITH THE USE OF THE MATERIALS, INCLUDING BUT NOT LIMITED TO THE RISKS OF PROGRAM ERRORS, DAMAGE TO OR LOSS OF DATA, PROGRAMS OR EQUIPMENT, AND UNAVAILABILITY OR INTERRUPTION OF OPERATIONS.  Some jurisdictions do not allow for the exclusion or limitation of implied warranties, so the above limitations or exclusions may not apply to You.
IN NO EVENT SHALL AMD OR ITS DIRECTORS, OFFICERS, EMPLOYEES AND AGENTS, ITS SUPPLIERS OR ITS LICENSORS BE LIABLE TO YOU OR ANY THIRD PARTIES IN RECEIPT OF THE MATERIALS UNDER ANY THEORY OF LIABILITY, WHETHER EQUITABLE, LEGAL OR COMMON LAW ACTION ARISING HEREUNDER FOR CONTRACT, STRICT LIABILITY, INDEMNITY, TORT (INCLUDING NEGLIGENCE), OR OTHERWISE FOR DAMAGES WHICH, IN THE AGGREGATE EXCEED TEN DOLLARS ($10.00).  IN NO EVENT SHALL AMD BE LIABLE FOR ANY CONSEQUENTIAL, INCIDENTAL, PUNITIVE OR SPECIAL DAMAGES, INCLUDING, BUT NOT LIMITED TO LOSS OF PROFITS, BUSINESS INTERRUPTION, OR LOSS OF INFORMATION ARISING OUT OF THE USE OF OR INABILITY TO USE THE MATERIALS, EVEN IF AMD HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.  BY USING THE MATERIALS WITHOUT CHARGE, YOU ACCEPT THIS ALLOCATION OF RISK.  Because some jurisdictions prohibit the exclusion or limitation of liability for consequential or incidental damages, the above limitation may not apply to you.
AMD does not assume any responsibility for any errors which may appear in the Materials or any other related information provided to you by AMD, or result from use of the Materials or any related information.  
You agree that you will not reverse engineer or decompile the Materials, in whole or in part, except as allowed by applicable law.
NO SUPPORT OBLIGATION: AMD is not obligated to furnish, support, or make any further information, software, technical information, know-how, or show-how available to you.  Additionally, AMD retains the right to modify the Materials at any time, without notice, and is not obligated to provide such modified Materials to you.
U.S. GOVERNMENT RESTRICTED RIGHTS: The Materials are provided with "RESTRICTED RIGHTS." Use, duplication, or disclosure by the Government is subject to the restrictions as set forth in FAR 52.227-14 and DFAR252.227-7013, et seq., or its successor.  Use of the Materials by the Government constitutes acknowledgement of AMD's proprietary rights in them.
EXPORT RESTRICTIONS: You shall adhere to all applicable U.S., European, and other export laws, including but not limited to the U.S. Export Administration Regulations (EAR), (15 C.F.R. Sections 730 through 774), and E.U. Council Regulation (EC) No 1334/2000 of 22 June 2000.  Further, pursuant to Section 740.6 of the EAR, you hereby certify that, except pursuant to a license granted by the United States Department of Commerce Bureau of Industry and Security or as otherwise permitted pursuant to a License Exception under the EAR, you will not (1) export, re-export or release to a national of a country in Country Groups D:1, E:1 or E:2 any restricted technology, software, or source code it receives from AMD, or (2) export to Country Groups D:1, E:1 or E:2 the direct product of such technology or software, if such foreign produced direct product is subject to national security controls as identified on the Commerce Control List (currently found in Supplement 1 to Part 774 of EAR).  For the most current Country Group listings, or for additional information about the EAR or Your obligations under those regulations, please refer to the U.S. Bureau of Industry and Securitys website at http://www.bis.doc.gov/.

----------------------------------------------------------------------------------------------------*/

#include "DXUT.h"
#include "SDKmisc.h"
#include "AMD_HD3D.h"

#ifndef SAFE_RELEASE
#define SAFE_RELEASE(p)      { if (p) { (p)->Release(); (p)=NULL; } }
#endif

AMD_HD3D g_HD3D;


// For drawing the pixel ID
struct StereoPixelIDVertex
{
	D3DXVECTOR3	v3Position;
	D3DXCOLOR	Color;
};

// Constant buffer layout for rendering PixelID
struct CB_PIXELID_CONSTANTS
{
    float    width;		// screen width 
    float    height;	// screen height
	float	 pad0;
	float	 pad1;
};



/*----------------------------------------------------------------------------------------------------
	AMD_HD3D Constructor
----------------------------------------------------------------------------------------------------*/
AMD_HD3D::AMD_HD3D()
{
    pStereo = NULL;
    enableStereo = true;
    pixelID = true;
	lineOffset = 0;
	width = 0;
	height = 0;
    pQuadBufferView = NULL;
    pQuadDepthStencilView = NULL;
    pRightEyeRT = NULL;
    pRightEyeView = NULL;
	pRightEyeSRV = NULL;
    pLeftEyeRT = NULL;
    pLeftEyeView = NULL;
	pLeftEyeSRV = NULL;
    pRightEyeDepth = NULL;
    pRightDepthView = NULL;
    pLeftEyeDepth = NULL;
    pLeftDepthView = NULL;
	pSingleSampleRT = NULL;
	pExtension = NULL;
	pStereoPixelIDVB = NULL;
	pStereoPixelIDVertexLayout = NULL;
	pStereoPixelIDVS = NULL;
	pStereoPixelIDPS = NULL;
	pcbPixelIDConstants = NULL;
}

/*----------------------------------------------------------------------------------------------------
	AMD_HD3D Destructor
----------------------------------------------------------------------------------------------------*/
AMD_HD3D::~AMD_HD3D()
{
	Release();
}

/*----------------------------------------------------------------------------------------------------
	AMD_HD3D OpenStereoInterface
----------------------------------------------------------------------------------------------------*/
HRESULT AMD_HD3D::OpenStereoInterface(ID3D11Device *pD3DDevice)
{
    PFNAmdDxExtCreate11 AmdDxExtCreate;
    HMODULE hDLL;
    HRESULT hr = S_OK;

#if defined _WIN64
    hDLL = GetModuleHandle("atidxx64.dll");
#else
    hDLL = GetModuleHandle(L"atidxx32.dll");
#endif

    // Find the DLL entry point
    AmdDxExtCreate = reinterpret_cast<PFNAmdDxExtCreate11>(GetProcAddress(hDLL, "AmdDxExtCreate11"));
    if (AmdDxExtCreate == NULL)
    {
        hr = E_FAIL;
    }
    // Create the extension object
    if (SUCCEEDED(hr))
    {
        hr = AmdDxExtCreate(pD3DDevice, &pExtension);
    }
    if (SUCCEEDED(hr))
    {
        pStereo = static_cast<IAmdDxExtQuadBufferStereo*>(pExtension->GetExtInterface(AmdDxExtQuadBufferStereoID));
		if (pStereo == NULL)
			hr = E_FAIL;
    }

	if (FAILED(hr))
	{
		CloseStereoInterface();
	}

    return hr;
}

/*----------------------------------------------------------------------------------------------------
	AMD_HD3D CloseStereoInterface
----------------------------------------------------------------------------------------------------*/
VOID AMD_HD3D::CloseStereoInterface()
{
    if (pStereo != NULL)
    {
        pStereo->Release();
        pStereo = NULL;
    }

	if (pExtension != NULL)
	{
        pExtension->Release();
		pExtension = NULL;
	}
}

/*----------------------------------------------------------------------------------------------------
	AMD_HD3D CreateStereoSwapChain
----------------------------------------------------------------------------------------------------*/
HRESULT AMD_HD3D::CreateStereoSwapChain( IDXGIFactory1* pIDXGIFactory, ID3D11Device* pd3d11Device, 
                                    DXGI_SWAP_CHAIN_DESC *pRequestedSwapChainDesc, 
                                    IDXGISwapChain **ppSwapChain )
{
    HRESULT hr;

	enableStereo = true;

	if (pRequestedSwapChainDesc->Windowed)
	{
		enableStereo = false;
		return E_FAIL;
	}

    if (pStereo == NULL)
    {   // try creating a stereo interface
        hr = OpenStereoInterface(pd3d11Device);
        if (FAILED(hr))
        {   // can't create a stereo interface - disable stereo
            enableStereo = false;
            return hr;
        }
    }

    // Enable quad buffer support for swap chain
    hr = pStereo->EnableQuadBufferStereo(TRUE);
    if (FAILED(hr))
    {   // can't create a stereo interface - disable stereo
        enableStereo = false;
		Release();
        return hr;
    }

    // Find a match with the requested swap chain description and the valid stereo ones
	UINT numModes = 0;
	int refreshMatch = -1;
	int resFormatMatch = -1;

	DXGI_MODE_DESC requestedModeDesc = pRequestedSwapChainDesc->BufferDesc;
	hr = pStereo->GetDisplayModeList(requestedModeDesc.Format, 0, &numModes, NULL);
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		Release();
		enableStereo = false;
		return hr;
	}

	if (numModes == 0)
	{
		pStereo->EnableQuadBufferStereo(FALSE);
        enableStereo = false;
		Release();
		return E_FAIL;
	}

	DXGI_MODE_DESC* pModeDescs = new DXGI_MODE_DESC[numModes];
	if (!pModeDescs)
	{
		pStereo->EnableQuadBufferStereo(FALSE);
        enableStereo = false;
		Release();
		return E_FAIL;
	}

	// enumerate all of the modes with the requestedModeDesc format
	hr = pStereo->GetDisplayModeList(requestedModeDesc.Format, 0, &numModes, pModeDescs);
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		delete[] pModeDescs;
        enableStereo = false;
		Release();
		return hr;
	}

	for (UINT i=0; i < numModes; ++i)
	{
		const DXGI_MODE_DESC& modeDesc = pModeDescs[i];
		if (modeDesc.Width != requestedModeDesc.Width)
			continue;
		if (modeDesc.Height != requestedModeDesc.Height)
			continue;
		if (modeDesc.Format != requestedModeDesc.Format)
			continue;
		// if it made it this far, the selected screen resolution and format match one of the possible stereo modes
		resFormatMatch = i;
		if ((modeDesc.RefreshRate.Numerator == requestedModeDesc.RefreshRate.Numerator)
			&& (modeDesc.RefreshRate.Denominator == requestedModeDesc.RefreshRate.Denominator))
		{
			refreshMatch = i; // found a match with the current refresh rate
			break;
		}
	}

	if (resFormatMatch < 0 )
	{
		// no valid stereo formats that match
		pStereo->EnableQuadBufferStereo(FALSE);
        enableStereo = false;
 		delete[] pModeDescs;
		Release();
        return E_FAIL;
	}
	int modeIndex = resFormatMatch;

	// Ideally we want the refresh rate to match what was requested. However, we'll 
	// automatically use a valid refresh rate if an invalid one was requested.
	if (refreshMatch >= 0)
	{
		modeIndex = refreshMatch;
	}

	// update the requested swap chain description with the actual one
	pRequestedSwapChainDesc->BufferDesc = pModeDescs[modeIndex];
 	delete[] pModeDescs;

	// always create a single sample swap chain
	requestedSampleDesc = pRequestedSwapChainDesc->SampleDesc;
	pRequestedSwapChainDesc->SampleDesc.Count = 1;

    // Create quad buffer swap chain
	hr = pIDXGIFactory->CreateSwapChain( pd3d11Device, pRequestedSwapChainDesc, ppSwapChain );
	DXUT_SetDebugName( *ppSwapChain, "HD3D: Swap Chain" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        enableStereo = false;
		Release();
        return hr;
	}
	pStereoSwapChain = *ppSwapChain;

    // Store line offset to right eye
	lineOffset = pStereo->GetLineOffset(*ppSwapChain);
	// A zero lineOffset indicates the quad buffer is not available for this configuration.
	if (lineOffset == 0)
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        enableStereo = false;
		Release();
        return E_FAIL;
	}

	width = pRequestedSwapChainDesc->BufferDesc.Width;
	height = pRequestedSwapChainDesc->BufferDesc.Height;

	// Create the right eye render target
	D3D11_TEXTURE2D_DESC textureDesc;
	textureDesc.ArraySize = 1;
	textureDesc.BindFlags = D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE;
	textureDesc.CPUAccessFlags = 0;
	textureDesc.Format = pRequestedSwapChainDesc->BufferDesc.Format;
	textureDesc.Width = width;
	textureDesc.Height = height;
	textureDesc.MipLevels = 1;
	textureDesc.SampleDesc = requestedSampleDesc;  // use the originally requested msaa setting
	textureDesc.Usage = D3D11_USAGE_DEFAULT;
	textureDesc.MiscFlags = 0;
	hr = pd3d11Device->CreateTexture2D(&textureDesc, NULL, &pRightEyeRT);
	DXUT_SetDebugName( pRightEyeRT, "HD3D: pRightEyeRT" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
 		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}
	hr = pd3d11Device->CreateRenderTargetView(pRightEyeRT, NULL, &pRightEyeView);
	DXUT_SetDebugName( pRightEyeView, "HD3D: pRightEyeView" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}
	hr = pd3d11Device->CreateShaderResourceView(pRightEyeRT, NULL, &pRightEyeSRV);
	DXUT_SetDebugName( pRightEyeSRV, "HD3D: pRightEyeSRV" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}

	// create the left eye render target
	hr = pd3d11Device->CreateTexture2D(&textureDesc, NULL, &pLeftEyeRT);
	DXUT_SetDebugName( pLeftEyeRT, "HD3D: pLeftEyeRT" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
 		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}
	hr = pd3d11Device->CreateRenderTargetView(pLeftEyeRT, NULL, &pLeftEyeView);
	DXUT_SetDebugName( pLeftEyeView, "HD3D: pLeftEyeView" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}
	hr = pd3d11Device->CreateShaderResourceView(pLeftEyeRT, NULL, &pLeftEyeSRV);
	DXUT_SetDebugName( pLeftEyeSRV, "HD3D: pLeftEyeSRV" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}

	// create an additional render target for MSAA
	if (requestedSampleDesc.Count > 1)
	{
		textureDesc.SampleDesc.Count = 1;  // need a single sample intermediate buffer
		hr = pd3d11Device->CreateTexture2D(&textureDesc, NULL, &pSingleSampleRT);
		DXUT_SetDebugName( pSingleSampleRT, "HD3D: pSingleSampleRT" );
		if (FAILED(hr))
		{
			pStereo->EnableQuadBufferStereo(FALSE);
 			pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
			pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
			Release();
			return hr;
		}
	}

	// create the depth buffers
	textureDesc.BindFlags = D3D10_BIND_DEPTH_STENCIL;
	textureDesc.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;
	textureDesc.SampleDesc = requestedSampleDesc;
	
	// right depth
	hr = pd3d11Device->CreateTexture2D(&textureDesc, NULL, &pRightEyeDepth);
	DXUT_SetDebugName( pRightEyeDepth, "HD3D: pRightEyeDepth" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
 		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}
	hr = pd3d11Device->CreateDepthStencilView(pRightEyeDepth, NULL, &pRightDepthView);
	DXUT_SetDebugName( pRightDepthView, "HD3D: pRightDepthView" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}

	// left depth
	hr = pd3d11Device->CreateTexture2D(&textureDesc, NULL, &pLeftEyeDepth);
	DXUT_SetDebugName( pLeftEyeDepth, "HD3D: pLeftEyeDepth" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
 		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}
	hr = pd3d11Device->CreateDepthStencilView(pRightEyeDepth, NULL, &pLeftDepthView);
	DXUT_SetDebugName( pLeftDepthView, "HD3D: pLeftDepthView" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}


	// create the pixel ID vertex buffer
	StereoPixelIDVertex VB[4];
    VB[0].v3Position.x = 1.0f;
	VB[0].v3Position.y = 1.0f;
    VB[0].v3Position.z = 0.0f;
	VB[0].Color = D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f );
    VB[1].v3Position.x = 2.0f;
	VB[1].v3Position.y = 1.0f;
    VB[1].v3Position.z = 0.0f;
	VB[1].Color = D3DXCOLOR( 0.0f, 0.0f, 0.0f, 1.0f );
    VB[2].v3Position.x = 1.0f;
	VB[2].v3Position.y = 1.0f;
    VB[2].v3Position.z = 0.0f;
	VB[2].Color = D3DXCOLOR( 0.0f, 0.0f, 0.0f, 1.0f );
    VB[3].v3Position.x = 2.0f;
	VB[3].v3Position.y = 1.0f;
    VB[3].v3Position.z = 0.0f;
	VB[3].Color = D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f );

	D3D11_BUFFER_DESC BD;
    BD.Usage = D3D11_USAGE_DYNAMIC;
    BD.ByteWidth = sizeof( StereoPixelIDVertex ) * 4;
    BD.BindFlags = D3D11_BIND_VERTEX_BUFFER;
    BD.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
    BD.MiscFlags = 0;

    D3D11_SUBRESOURCE_DATA InitData;
    InitData.SysMemPitch = 0;
    InitData.SysMemSlicePitch = 0;
    InitData.pSysMem = VB;

    hr = pd3d11Device->CreateBuffer( &BD, &InitData, &pStereoPixelIDVB );
	DXUT_SetDebugName( pStereoPixelIDVB, "HD3D: pStereoPixelIDVB" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}

    // Create stereo PixelID shaders and input layout

	ID3DBlob* pBlob = NULL;
	ID3DBlob* pErrorBlob = NULL;
	DWORD dwShaderFlags = D3DCOMPILE_ENABLE_STRICTNESS;
	WCHAR ShaderFileStr[MAX_PATH];

	DXUTFindDXSDKMediaFileCch( ShaderFileStr, MAX_PATH, L"AMD_HD3D.hlsl" );

	// VS
    hr = D3DX11CompileFromFile( ShaderFileStr, NULL, NULL, "VS_RenderStereoPixelID", "vs_4_0", 
		dwShaderFlags, 0, NULL, &pBlob, &pErrorBlob, NULL );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        SAFE_RELEASE( pBlob );
		SAFE_RELEASE( pErrorBlob );
		SAFE_RELEASE( pBlob );
        Release();
        return hr;
	}
	hr = pd3d11Device->CreateVertexShader( pBlob->GetBufferPointer(), pBlob->GetBufferSize(), NULL, &pStereoPixelIDVS );
	DXUT_SetDebugName( pStereoPixelIDVS, "HD3D: pStereoPixelIDVS" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        SAFE_RELEASE( pBlob );
		SAFE_RELEASE( pErrorBlob );
		SAFE_RELEASE( pBlob );
        Release();
        return hr;
	}
	// Input layout
	const D3D11_INPUT_ELEMENT_DESC StereoPixelIDLayout[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
        { "COLOR", 0, DXGI_FORMAT_R32G32B32A32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0 },
    };
	hr = pd3d11Device->CreateInputLayout( StereoPixelIDLayout, ARRAYSIZE( StereoPixelIDLayout ), pBlob->GetBufferPointer(),
												pBlob->GetBufferSize(), &pStereoPixelIDVertexLayout );
	DXUT_SetDebugName( pStereoPixelIDVertexLayout, "HD3D: pStereoPixelIDVertexLayout" );
    SAFE_RELEASE( pBlob );
	SAFE_RELEASE( pErrorBlob );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
		SAFE_RELEASE( pBlob );
        Release();
        return hr;
	}

	// PS
    hr = D3DX11CompileFromFile( ShaderFileStr, NULL, NULL, "PS_RenderStereoPixelID", "ps_4_0", 
		dwShaderFlags, 0, NULL, &pBlob, &pErrorBlob, NULL );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
		SAFE_RELEASE( pBlob );
		SAFE_RELEASE( pErrorBlob );
        Release();
        return hr;
	}
	hr = pd3d11Device->CreatePixelShader( pBlob->GetBufferPointer(), pBlob->GetBufferSize(), NULL, &pStereoPixelIDPS );
	DXUT_SetDebugName( pStereoPixelIDPS, "HD3D: pStereoPixelIDPS" );
    SAFE_RELEASE( pBlob );
	SAFE_RELEASE( pErrorBlob );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
        Release();
        return hr;
	}

	// create the PixelID constant buffer
    D3D11_BUFFER_DESC Desc;
	Desc.Usage = D3D11_USAGE_IMMUTABLE;
    Desc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    Desc.CPUAccessFlags = 0;
    Desc.MiscFlags = 0;    
    Desc.ByteWidth = sizeof( CB_PIXELID_CONSTANTS );

	CB_PIXELID_CONSTANTS constantData;
	constantData.width = (float)width;
	constantData.height = (float)height;

    InitData.pSysMem = &constantData;
	InitData.SysMemPitch = 0;
	InitData.SysMemSlicePitch = 0;

    hr = pd3d11Device->CreateBuffer( &Desc, &InitData, &pcbPixelIDConstants );
	DXUT_SetDebugName( pcbPixelIDConstants, "HD3D: pcbPixelIDConstants" );
	if (FAILED(hr))
	{
		pStereo->EnableQuadBufferStereo(FALSE);
		pRequestedSwapChainDesc->BufferDesc = requestedModeDesc;
		pRequestedSwapChainDesc->SampleDesc = requestedSampleDesc;
		SAFE_RELEASE( pBlob );
        Release();
        return hr;
	}

	return S_OK;

}

/*----------------------------------------------------------------------------------------------------
	GetBackBufferView
----------------------------------------------------------------------------------------------------*/
void AMD_HD3D::InitBackBufferView(ID3D11RenderTargetView *pRenderTargetView, 
								 ID3D11DepthStencilView *pDepthStencilView)
{
	if (enableStereo)
	{
		pQuadBufferView = pRenderTargetView;
		pQuadDepthStencilView = pDepthStencilView;
	}
}

/*----------------------------------------------------------------------------------------------------
	SwitchToRightEye
----------------------------------------------------------------------------------------------------*/
void AMD_HD3D::SwitchToRightEye(ID3D11DeviceContext *pdeviceContext)
{
	if (!enableStereo)
		return;

	pdeviceContext->OMSetRenderTargets(1, &pRightEyeView, pRightDepthView);
}

/*----------------------------------------------------------------------------------------------------
	SwitchToLeftEye
----------------------------------------------------------------------------------------------------*/
void AMD_HD3D::SwitchToLeftEye(ID3D11DeviceContext *pdeviceContext)
{
	if (!enableStereo)
		return;

	pdeviceContext->OMSetRenderTargets(1, &pLeftEyeView, pLeftDepthView);
}

/*----------------------------------------------------------------------------------------------------
	ClearRenderTargets - clear the left and right offscreen render targets
----------------------------------------------------------------------------------------------------*/
void AMD_HD3D::ClearRenderTargets(ID3D11DeviceContext *pdeviceContext)
{
	if (!enableStereo)
		return;

    float ClearColor[4] = { 0.0f, 0.25f, 0.25f, 0.55f };
	pdeviceContext->ClearRenderTargetView( pLeftEyeView, ClearColor );
	pdeviceContext->ClearDepthStencilView( pLeftDepthView, D3D11_CLEAR_DEPTH, 1.0, 0 );
	pdeviceContext->ClearRenderTargetView( pRightEyeView, ClearColor );
	pdeviceContext->ClearDepthStencilView( pRightDepthView, D3D11_CLEAR_DEPTH, 1.0, 0 );
}


/*----------------------------------------------------------------------------------------------------
	UpdateQuadBuffer
----------------------------------------------------------------------------------------------------*/
void AMD_HD3D::UpdateQuadBuffer(ID3D11DeviceContext *pdeviceContext)
{
	if (!enableStereo)
		return;

	ID3D11Texture2D *p_RT;
	pStereoSwapChain->GetBuffer(0, __uuidof(p_RT), reinterpret_cast<void**>(&p_RT));

	// copy the right and left render targets into the stereo swap chain
	DXGI_SWAP_CHAIN_DESC swapChainDesc;
	pStereoSwapChain->GetDesc(&swapChainDesc);
	if (requestedSampleDesc.Count > 1)
	{	// multisampled
		pdeviceContext->ResolveSubresource(pSingleSampleRT, 0, pRightEyeRT, 0, 
			swapChainDesc.BufferDesc.Format);
		pdeviceContext->CopySubresourceRegion(p_RT, 0, 0, lineOffset, 0, pSingleSampleRT, 0, NULL);
		pdeviceContext->ResolveSubresource(pSingleSampleRT, 0, pLeftEyeRT, 0, 
			swapChainDesc.BufferDesc.Format);
		pdeviceContext->CopySubresourceRegion(p_RT, 0, 0, 0, 0, pSingleSampleRT, 0, NULL);
	}
	else
	{	// single sampled
		pdeviceContext->CopySubresourceRegion(p_RT, 0, 0, lineOffset, 0, pRightEyeRT, 0, NULL);
		pdeviceContext->CopySubresourceRegion(p_RT, 0, 0, 0, 0, pLeftEyeRT, 0, NULL);
	}
	p_RT->Release();
}

/*----------------------------------------------------------------------------------------------------
	RenderPixelID
----------------------------------------------------------------------------------------------------*/
void AMD_HD3D::RenderPixelID(ID3D11DeviceContext *pdeviceContext)
{
	if (!enableStereo)
		return;

	if (!pixelID)
		return;

	// use the quad buffer render target
	pdeviceContext->OMSetRenderTargets(1, &pQuadBufferView, pQuadDepthStencilView);

    // Save the old viewport
    D3D11_VIEWPORT vpOld[D3D11_VIEWPORT_AND_SCISSORRECT_MAX_INDEX];
    UINT nViewPorts = 1;
    pdeviceContext->RSGetViewports( &nViewPorts, vpOld );

    D3D11_VIEWPORT vp;
	vp.Width = (float)width;
	vp.Height = (float)height;
	vp.MinDepth = 0.0f;
	vp.MaxDepth = 1.0f;
	vp.TopLeftX = 0.0f;
	vp.TopLeftY = 0.0f;
	pdeviceContext->RSSetViewports( 1, &vp );

    UINT Stride = sizeof( StereoPixelIDVertex );
	UINT Offset = 0;
    pdeviceContext->VSSetShader( pStereoPixelIDVS, NULL, 0 );
    pdeviceContext->PSSetShader( pStereoPixelIDPS, NULL, 0 );
	pdeviceContext->VSSetConstantBuffers( 0, 1, &pcbPixelIDConstants);
    pdeviceContext->IASetInputLayout( pStereoPixelIDVertexLayout );
    pdeviceContext->IASetVertexBuffers( 0, 1, &pStereoPixelIDVB, &Stride, &Offset );
    pdeviceContext->IASetPrimitiveTopology( D3D11_PRIMITIVE_TOPOLOGY_POINTLIST );
    pdeviceContext->Draw( 2, 0 );

    vp.TopLeftY = (float)lineOffset;
    pdeviceContext->RSSetViewports( 1, &vp );
    pdeviceContext->Draw( 2, 2 );

    pdeviceContext->VSSetShader( NULL, NULL, 0 );
    pdeviceContext->PSSetShader( NULL, NULL, 0 );

    // Restore the Old viewport
    pdeviceContext->RSSetViewports( nViewPorts, vpOld );
}

/*----------------------------------------------------------------------------------------------------
	Release
----------------------------------------------------------------------------------------------------*/
void AMD_HD3D::Release()
{
    SAFE_RELEASE(pRightEyeView);
	SAFE_RELEASE(pRightEyeSRV);
    SAFE_RELEASE(pRightEyeRT);
    SAFE_RELEASE(pLeftEyeView);
	SAFE_RELEASE(pLeftEyeSRV);
    SAFE_RELEASE(pLeftEyeRT);
    SAFE_RELEASE(pRightDepthView);
    SAFE_RELEASE(pRightEyeDepth);
    SAFE_RELEASE(pLeftDepthView);
    SAFE_RELEASE(pLeftEyeDepth);
    SAFE_RELEASE(pSingleSampleRT);
	SAFE_RELEASE(pStereoPixelIDVB);
	SAFE_RELEASE(pStereoPixelIDVertexLayout);
	SAFE_RELEASE(pStereoPixelIDVS);
	SAFE_RELEASE(pStereoPixelIDPS);
	SAFE_RELEASE(pcbPixelIDConstants);
	CloseStereoInterface();

	enableStereo = false;
}
