/*----------------------------------------------------------------------------------------------------
	AMD_Stereo.cpp

	AMD Quad Buffer Stereo Class

Copyright  2011 ADVANCED MICRO DEVICES, INC.  All Rights Reserved.

AMD is granting you permission to use this software and documentation (if any) (collectively, the Materials) pursuant to the terms and conditions of the Software Development Kit License Agreement included with the Materials.  This header does NOT give you permission to use the Materials or any rights under AMDs intellectual property.  Your use of any portion of these Materials shall constitute your acceptance of the terms and conditions of such Software Development Kit License Agreement.  If you do not agree to the terms and conditions of the Software Development Kit License Agreement, you do not have permission to use any portion of these Materials.  If you do not have a copy of the Software Development Kit License Agreement, contact your AMD representative for a copy.
CONFIDENTIALITY:  The Materials and all other related information, identified as confidential and provided to you by AMD shall be kept confidential in accordance with the terms and conditions of the Software Development Kit License Agreement.  If no Software Development Kit License Agreement exists between you and AMD, you agree to keep these Materials confidential and to not disclose it to any third party.
LIMITATION OF LIABILITY: THE MATERIALS ARE PROVIDED AS IS WITHOUT ANY EXPRESS OR IMPLIED WARRANTY OF ANY KIND INCLUDING WARRANTIES OF MERCHANTABILITY, NONINFRINGEMENT OF THIRD-PARTY INTELLECTUAL PROPERTY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE, OR THOSE ARISING FROM CUSTOM OF TRADE OR COURSE OF USAGE.  
FOR CLARIFICATION, THE ENTIRE RISK ARISING OUT OF USE OR PERFORMANCE OF THE MATERIALS REMAINS WITH YOU.  AMD DOES NOT WARRANT, GUARANTEE, OR MAKE ANY REPRESENTATIONS AS TO THE CORRECTNESS, ACCURACY, COMPLETENESS, QUALITY, OR RELIABILITY OF THE MATERIALS.  AMD DOES NOT WARRANT THAT OPERATION OF THE MATERIALS WILL BE UNINTERRUPTED OR ERROR-FREE.  YOU ARE RESPONSIBLE FOR DETERMINING THE APPROPRIATENESS OF USING THE SOFTWARE AND ASSUME ALL RISKS ASSOCIATED WITH THE USE OF THE MATERIALS, INCLUDING BUT NOT LIMITED TO THE RISKS OF PROGRAM ERRORS, DAMAGE TO OR LOSS OF DATA, PROGRAMS OR EQUIPMENT, AND UNAVAILABILITY OR INTERRUPTION OF OPERATIONS.  Some jurisdictions do not allow for the exclusion or limitation of implied warranties, so the above limitations or exclusions may not apply to You.
IN NO EVENT SHALL AMD OR ITS DIRECTORS, OFFICERS, EMPLOYEES AND AGENTS, ITS SUPPLIERS OR ITS LICENSORS BE LIABLE TO YOU OR ANY THIRD PARTIES IN RECEIPT OF THE MATERIALS UNDER ANY THEORY OF LIABILITY, WHETHER EQUITABLE, LEGAL OR COMMON LAW ACTION ARISING HEREUNDER FOR CONTRACT, STRICT LIABILITY, INDEMNITY, TORT (INCLUDING NEGLIGENCE), OR OTHERWISE FOR DAMAGES WHICH, IN THE AGGREGATE EXCEED TEN DOLLARS ($10.00).  IN NO EVENT SHALL AMD BE LIABLE FOR ANY CONSEQUENTIAL, INCIDENTAL, PUNITIVE OR SPECIAL DAMAGES, INCLUDING, BUT NOT LIMITED TO LOSS OF PROFITS, BUSINESS INTERRUPTION, OR LOSS OF INFORMATION ARISING OUT OF THE USE OF OR INABILITY TO USE THE MATERIALS, EVEN IF AMD HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.  BY USING THE MATERIALS WITHOUT CHARGE, YOU ACCEPT THIS ALLOCATION OF RISK.  Because some jurisdictions prohibit the exclusion or limitation of liability for consequential or incidental damages, the above limitation may not apply to you.
AMD does not assume any responsibility for any errors which may appear in the Materials or any other related information provided to you by AMD, or result from use of the Materials or any related information.  
You agree that you will not reverse engineer or decompile the Materials, in whole or in part, except as allowed by applicable law.
NO SUPPORT OBLIGATION: AMD is not obligated to furnish, support, or make any further information, software, technical information, know-how, or show-how available to you.  Additionally, AMD retains the right to modify the Materials at any time, without notice, and is not obligated to provide such modified Materials to you.
U.S. GOVERNMENT RESTRICTED RIGHTS: The Materials are provided with "RESTRICTED RIGHTS." Use, duplication, or disclosure by the Government is subject to the restrictions as set forth in FAR 52.227-14 and DFAR252.227-7013, et seq., or its successor.  Use of the Materials by the Government constitutes acknowledgement of AMD's proprietary rights in them.
EXPORT RESTRICTIONS: You shall adhere to all applicable U.S., European, and other export laws, including but not limited to the U.S. Export Administration Regulations (EAR), (15 C.F.R. Sections 730 through 774), and E.U. Council Regulation (EC) No 1334/2000 of 22 June 2000.  Further, pursuant to Section 740.6 of the EAR, you hereby certify that, except pursuant to a license granted by the United States Department of Commerce Bureau of Industry and Security or as otherwise permitted pursuant to a License Exception under the EAR, you will not (1) export, re-export or release to a national of a country in Country Groups D:1, E:1 or E:2 any restricted technology, software, or source code it receives from AMD, or (2) export to Country Groups D:1, E:1 or E:2 the direct product of such technology or software, if such foreign produced direct product is subject to national security controls as identified on the Commerce Control List (currently found in Supplement 1 to Part 774 of EAR).  For the most current Country Group listings, or for additional information about the EAR or Your obligations under those regulations, please refer to the U.S. Bureau of Industry and Securitys website at http://www.bis.doc.gov/.

----------------------------------------------------------------------------------------------------*/

#include "AMD_Stereo.h"
#include <iostream>

AMD_Stereo g_AMDstereo;

/*----------------------------------------------------------------------------------------------------
	AMD_Stereo Constructor
----------------------------------------------------------------------------------------------------*/
AMD_Stereo::AMD_Stereo()
{
	pixelID = true;
	enableStereo = true;
	pd3dDevice = NULL;
	pBackBufferDepthStencil = NULL;
	pDriverComSurface = NULL;
	pRightEyeRT = NULL;
}

/*----------------------------------------------------------------------------------------------------
	AMD_Stereo Destructor
----------------------------------------------------------------------------------------------------*/
AMD_Stereo::~AMD_Stereo()
{
	Release();
}

/*----------------------------------------------------------------------------------------------------
	CreateStereoDevice
----------------------------------------------------------------------------------------------------*/
HRESULT AMD_Stereo::CreateStereoDevice(IDirect3D9* pD3D, UINT adapter, D3DDEVTYPE type, HWND hWindow, 
									  DWORD flags, D3DPRESENT_PARAMETERS *pParams, 
									  IDirect3DDevice9 **ppDevice)
{

	if (pParams->Windowed)
	{
		// stereo requires windowed mode
		enableStereo = false;
		return E_FAIL;
	}

	enableStereo = true;

	// save the requested MSAA settings
	multiSampleType = pParams->MultiSampleType;
	multiSampleQuality = pParams->MultiSampleQuality;

	// First create a temporary windowed device
	D3DPRESENT_PARAMETERS tmpParams = *pParams;
	tmpParams.Windowed = true;
	tmpParams.EnableAutoDepthStencil = false;
	tmpParams.Flags = 0;
	tmpParams.SwapEffect = D3DSWAPEFFECT_COPY;
	tmpParams.PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;
	tmpParams.BackBufferWidth = 2;
	tmpParams.BackBufferHeight = 2;
	tmpParams.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
	tmpParams.MultiSampleType = D3DMULTISAMPLE_NONE;
	tmpParams.BackBufferCount = 1;

	HRESULT hr;
    hr = pD3D->CreateDevice( adapter, type,hWindow, flags, &tmpParams, &pd3dDevice );
	if( FAILED( hr ) )
    {
		enableStereo = false;
        return hr;
    }

	//Create a surface to be used to communicate with the driver
	hr = CreateCommSurface();
	if( FAILED( hr ) )
    {
		pd3dDevice->Release();
		enableStereo = false;
        return hr;
    }
	//Send the command to the driver using the temporary surface
	hr = SendStereoCommand(ATI_STEREO_ENABLESTEREO, NULL, 0, 0, 0);
	if( FAILED( hr ) )
    {
		pd3dDevice->Release();
		pDriverComSurface->Release();
		pDriverComSurface = NULL;
		enableStereo = false;
        return hr;
    }
	// See what stereo modes are available
	ATIDX9GETDISPLAYMODES displayModeParams;
	displayModeParams.dwNumModes    = 0;
	displayModeParams.pStereoModes = NULL;

	//Send stereo command to get the number of available stereo modes.
	hr = SendStereoCommand(ATI_STEREO_GETDISPLAYMODES, (BYTE *)(&displayModeParams),
						  sizeof(ATIDX9GETDISPLAYMODES), 0, 0);  
	if( FAILED( hr ) )
    {
		pd3dDevice->Release();
		pDriverComSurface->Release();
		pDriverComSurface = NULL;
		enableStereo = false;
        return hr;
    }

	if(displayModeParams.dwNumModes != 0)
	{
	   //Allocating memory to get the list of modes.
	   displayModeParams.pStereoModes = new D3DDISPLAYMODE[displayModeParams.dwNumModes];
	   
	   //Send stereo command to get the list of stereo modes
	   hr = SendStereoCommand(ATI_STEREO_GETDISPLAYMODES, (BYTE *)(&displayModeParams), 
							 sizeof(ATIDX9GETDISPLAYMODES), 0, 0);
		if( FAILED( hr ) )
		{
			pd3dDevice->Release();
			pDriverComSurface->Release();
			pDriverComSurface = NULL;
			enableStereo = false;
			delete[] displayModeParams.pStereoModes;
			return hr;
		}
	}

	D3DDISPLAYMODE currentMode;
	pd3dDevice->GetDisplayMode(0, &currentMode);

	int refreshMatch = -1;
	int resFormatMatch = -1;
	for (int i = 0; i < (int)displayModeParams.dwNumModes; i++)
	{
		if (displayModeParams.pStereoModes[i].Width != pParams->BackBufferWidth)
			continue;
		if (displayModeParams.pStereoModes[i].Height != pParams->BackBufferHeight)
			continue;
		if ( (displayModeParams.pStereoModes[i].Format != pParams->BackBufferFormat)
			&& ((displayModeParams.pStereoModes[i].Format != D3DFMT_X8R8G8B8) || (pParams->BackBufferFormat != D3DFMT_A8R8G8B8)))
			continue;
		// if it made it this far, the selected screen resolution and format match one of the possible stereo modes
		resFormatMatch = i;
		if (displayModeParams.pStereoModes[i].RefreshRate == currentMode.RefreshRate)
		{
			refreshMatch = i; // found a match with the current refresh
			break;
		}
	}

	ReleaseCommSurface();

	if (resFormatMatch < 0 )
	{
		enableStereo = false;
		pd3dDevice->Release();
		delete[] displayModeParams.pStereoModes;
		return -1;
	}
	else
	{
		int displayModeIndex;
		if (refreshMatch >= 0)
		{	// the current refresh rate matches a valid stereo mode
			displayModeIndex = refreshMatch;
		}
		else
		{	// the current refresh rate doesn't match, so try to change it with Reset()
			displayModeIndex = resFormatMatch;
		}

		ChangePresentParams(pParams);
		pParams->FullScreen_RefreshRateInHz = displayModeParams.pStereoModes[displayModeIndex].RefreshRate;;
		pParams->BackBufferFormat = displayModeParams.pStereoModes[displayModeIndex].Format;
		delete[] displayModeParams.pStereoModes;

		hr = pd3dDevice->Reset(pParams); // Call reset to create the full screen stereo quad buffers
		if( FAILED( hr ) )
		{
			pd3dDevice->Release();
			enableStereo = false;
			return hr;
		}
		
		pParams->MultiSampleType = multiSampleType;
		pParams->MultiSampleQuality = multiSampleQuality;
		params = *pParams;
		*ppDevice = pd3dDevice;
	}

	return hr;
}


/*----------------------------------------------------------------------------------------------------
	CreateCommSurface
----------------------------------------------------------------------------------------------------*/
HRESULT AMD_Stereo::CreateCommSurface()
{
	HRESULT hr;

	//Create a surface to be used to communicate with the driver
	hr = pd3dDevice->CreateOffscreenPlainSurface(10, 10, (D3DFORMAT)FOURCC_AQBS, D3DPOOL_DEFAULT, &pDriverComSurface, NULL);

	return hr;

}

/*----------------------------------------------------------------------------------------------------
	ReleaseCommSurface
----------------------------------------------------------------------------------------------------*/
void AMD_Stereo::ReleaseCommSurface()
{
	ULONG refCount;
	if (pDriverComSurface)
	{
		refCount = 1;
		while(refCount)
		{
			refCount = pDriverComSurface->Release();
		}
		pDriverComSurface = NULL;
	}
}

/*----------------------------------------------------------------------------------------------------
	CreateDefaultPoolResources
----------------------------------------------------------------------------------------------------*/
HRESULT AMD_Stereo::CreateDefaultPoolResources()
{
	HRESULT hr;

	if (!enableStereo)
		return S_OK;

	hr = pd3dDevice->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &pBackBufferSurface);
	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}

	hr = CreateCommSurface();
	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}

	//Retrieve the line offset
	hr = SendStereoCommand(ATI_STEREO_GETLINEOFFSET, (BYTE *)(&lineOffset), sizeof(DWORD), 0, 0);
	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}


	// see if lineOffset is valid
	if (lineOffset == 0)
	{ 
		// Something went wrong with the device reset. Allocate a depth buffer and disable stereo.
		enableStereo = false;
		hr = pd3dDevice->CreateDepthStencilSurface(params.BackBufferWidth, params.BackBufferHeight, params.AutoDepthStencilFormat,
			params.MultiSampleType, params.MultiSampleQuality, false, &pBackBufferDepthStencil, NULL);
		pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
		if( FAILED( hr ) )
		{
			return hr;
		}
		hr = pd3dDevice->SetDepthStencilSurface(pBackBufferDepthStencil);
		if( FAILED( hr ) )
		{
			return hr;
		}
		return S_OK;
	}

	// create an extra large depth buffer to handle both left and right buffers
	hr = pd3dDevice->CreateDepthStencilSurface(params.BackBufferWidth, 2 * lineOffset, params.AutoDepthStencilFormat,
		D3DMULTISAMPLE_2_SAMPLES, params.MultiSampleQuality, false, &pBackBufferDepthStencil, NULL);
	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}

	hr = pd3dDevice->SetDepthStencilSurface(pBackBufferDepthStencil);
	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}

	// Create the right eye texture
	hr = pd3dDevice->CreateRenderTarget( params.BackBufferWidth,
									params.BackBufferHeight,
									params.BackBufferFormat,
									multiSampleType,
									multiSampleQuality,
									false, &pRightEyeRT, NULL );

	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}

	// Create the right eye depth buffer
	hr = pd3dDevice->CreateDepthStencilSurface(params.BackBufferWidth,
									params.BackBufferHeight,
									D3DFMT_D24S8,
									multiSampleType, 
									multiSampleQuality,
									false, &pRightEyeDS, NULL);
	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}

	// Create the left eye texture
	hr = pd3dDevice->CreateRenderTarget( params.BackBufferWidth,
									params.BackBufferHeight,
									params.BackBufferFormat,
									multiSampleType,
									multiSampleQuality,
									false, &pLeftEyeRT, NULL );

	// Create the left eye depth buffer
	hr = pd3dDevice->CreateDepthStencilSurface(params.BackBufferWidth,
									params.BackBufferHeight,
									D3DFMT_D24S8,
									multiSampleType, 
									multiSampleQuality,
									false, &pLeftEyeDS, NULL);
	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}

	if( FAILED( hr ) )
	{
		enableStereo = false;
		return hr;
	}

	pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
	backBufferWidth = params.BackBufferWidth;
	backBufferHeight = params.BackBufferHeight;

	return S_OK;
}

/*----------------------------------------------------------------------------------------------------
	PrepareForReset
----------------------------------------------------------------------------------------------------*/
void AMD_Stereo::ChangePresentParams(D3DPRESENT_PARAMETERS *pParams)
{
	if (!enableStereo)
		return;

	if (pParams->Windowed)
		return;

	//A valid multisample value other then 0 or 1 must be set for stereo. (ex 2)
	pParams->MultiSampleType = D3DMULTISAMPLE_2_SAMPLES;
	pParams->Flags = 0; // can't lock the back buffer
	pParams->EnableAutoDepthStencil = false; // need to create a special depth buffer
}							  

/*----------------------------------------------------------------------------------------------------
	SendStereoCommand
----------------------------------------------------------------------------------------------------*/
HRESULT	AMD_Stereo::SendStereoCommand(ATIDX9STEREOCOMMAND stereoCommand, BYTE *pOutBuffer, 
									  DWORD dwOutBufferSize, BYTE *pInBuffer, 
									  DWORD dwInBufferSize)
{
	HRESULT hr;
	ATIDX9STEREOCOMMPACKET *pCommPacket;
	D3DLOCKED_RECT lockedRect;

	hr = pDriverComSurface->LockRect(&lockedRect, 0, 0);
	if(FAILED(hr))
	{
		return hr;
	}
	pCommPacket = (ATIDX9STEREOCOMMPACKET *)(lockedRect.pBits);
	pCommPacket->dwSignature = 'STER';
	pCommPacket->pResult = &hr;
	pCommPacket->stereoCommand = stereoCommand;
	if (pOutBuffer && !dwOutBufferSize)
	{
		return hr;
	}
	pCommPacket->pOutBuffer = pOutBuffer;
	pCommPacket->dwOutBufferSize = dwOutBufferSize;
	if (pInBuffer && !dwInBufferSize)
	{
		return hr;
	}
	pCommPacket->pInBuffer = pInBuffer;
	pCommPacket->dwInBufferSize = dwInBufferSize;
	pDriverComSurface->UnlockRect();
	return hr;
}


/*----------------------------------------------------------------------------------------------------
	RenderPixelID
----------------------------------------------------------------------------------------------------*/
HRESULT AMD_Stereo::RenderPixelID()
{
	if (!pixelID)
		return S_OK;

	struct Vertex
	{
		FLOAT		x, y, z, w;
		D3DCOLOR	diffuse;
	};
	
	Vertex verts[2];

	verts[0].x = 0;
	verts[0].z = 0.5f;
	verts[0].w = 1;
	verts[1].x = 1;
	verts[1].z = 0.5f;
	verts[1].w = 1;

	D3DVIEWPORT9 viewPort;
	viewPort.X = 0;
	viewPort.Width = backBufferWidth;
	viewPort.Height = backBufferHeight;
	viewPort.MinZ = 0;
	viewPort.MaxZ = 1;

	pd3dDevice->BeginScene();
	// save state
	DWORD zEnable;
	pd3dDevice->GetRenderState(D3DRS_ZENABLE, &zEnable);

	pd3dDevice->SetFVF(D3DFVF_XYZRHW | D3DFVF_DIFFUSE);
	D3DXMATRIX identityMatrix;
	D3DXMatrixIdentity(&identityMatrix);
	pd3dDevice->SetTransform(D3DTS_VIEW, &identityMatrix);
	pd3dDevice->SetTransform(D3DTS_WORLD, &identityMatrix);
	D3DXMATRIX orthoMatrix;
	D3DXMatrixOrthoRH(&orthoMatrix, (float)backBufferWidth, (float)backBufferHeight, 0.0, 1.0f);
	pd3dDevice->SetTransform(D3DTS_PROJECTION, &orthoMatrix);
	pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
	float pointSize = 1.0f;
	pd3dDevice->SetRenderTarget(0, pBackBufferSurface);
	pd3dDevice->SetRenderState(D3DRS_POINTSIZE, *((DWORD*)&pointSize));
	pd3dDevice->SetRenderState(D3DRS_POINTSCALEENABLE, FALSE);
	pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
	pd3dDevice->SetVertexShader(NULL);
	pd3dDevice->SetPixelShader(NULL);

	// draw right eye
	viewPort.Y = lineOffset;
	verts[0].y = (float)lineOffset;
	verts[1].y = (float)lineOffset;
	pd3dDevice->SetViewport(&viewPort);
	verts[0].diffuse = D3DCOLOR D3DCOLOR_ARGB(255, 0, 0, 0);
	verts[1].diffuse = D3DCOLOR D3DCOLOR_ARGB(255, 255, 255, 255);
	pd3dDevice->DrawPrimitiveUP(D3DPT_POINTLIST, 2, verts, sizeof(Vertex));

	// draw left eye
	viewPort.Y = 0;
	pd3dDevice->SetViewport(&viewPort);
	verts[0].y = 0;
	verts[1].y = 0;
	verts[0].diffuse = D3DCOLOR D3DCOLOR_ARGB(255, 255, 255, 255);
	verts[1].diffuse = D3DCOLOR D3DCOLOR_ARGB(255, 0, 0, 0);
	pd3dDevice->DrawPrimitiveUP(D3DPT_POINTLIST, 2, verts, sizeof(Vertex));

	// restore state
	pd3dDevice->SetRenderState(D3DRS_ZENABLE, zEnable);
	pd3dDevice->EndScene();
	return S_OK;
}

/*----------------------------------------------------------------------------------------------------
	SwitchToRightEye
----------------------------------------------------------------------------------------------------*/
void AMD_Stereo::SwitchToRightEye()
{
	pd3dDevice->SetRenderTarget(0, pRightEyeRT);
	pd3dDevice->SetDepthStencilSurface(pRightEyeDS);
}

/*----------------------------------------------------------------------------------------------------
	SwitchToLeftEye
----------------------------------------------------------------------------------------------------*/
void AMD_Stereo::SwitchToLeftEye()
{
	pd3dDevice->SetRenderTarget(0, pLeftEyeRT);
	pd3dDevice->SetDepthStencilSurface(pLeftEyeDS);
}

/*----------------------------------------------------------------------------------------------------
	UpdateQuadBuffer
----------------------------------------------------------------------------------------------------*/
void AMD_Stereo::UpdateQuadBuffer()
{
	// Use the quad buffer render target
	pd3dDevice->SetRenderTarget(0, pBackBufferSurface);

	// update the quad buffer with the right render target
	D3DVIEWPORT9 viewPort;
	viewPort.X = 0;
	viewPort.Y = lineOffset;
	viewPort.Width = backBufferWidth;
	viewPort.Height = backBufferHeight;
	viewPort.MinZ = 0;
	viewPort.MaxZ = 1;
	pd3dDevice->SetViewport(&viewPort);
	// set the right quad buffer as the destination for StretchRect
	DWORD dwEye = ATI_STEREO_RIGHTEYE;
	SendStereoCommand(ATI_STEREO_SETDSTEYE, NULL, 0, (BYTE *)&dwEye, sizeof(dwEye));
	pd3dDevice->StretchRect(pRightEyeRT, NULL, pBackBufferSurface, NULL, D3DTEXF_LINEAR);
	// set the left quad buffer as the destination for StretchRect
	viewPort.Y = 0;
	pd3dDevice->SetViewport(&viewPort);
	dwEye = ATI_STEREO_LEFTEYE;
	g_AMDstereo.SendStereoCommand(ATI_STEREO_SETDSTEYE, NULL, 0, (BYTE *)&dwEye, sizeof(dwEye));
	pd3dDevice->StretchRect(pLeftEyeRT, NULL, pBackBufferSurface, NULL, D3DTEXF_LINEAR);
}

/*----------------------------------------------------------------------------------------------------
	Release
----------------------------------------------------------------------------------------------------*/
void AMD_Stereo::Release()
{
	ULONG refCount;

	ReleaseCommSurface();

	if (pRightEyeRT)
	{
		refCount = 1;
		while(refCount)
		{
			refCount = pRightEyeRT->Release();
		}
	}
	pRightEyeRT = 0;

	if (pLeftEyeRT)
	{
		refCount = 1;
		while(refCount)
		{
			refCount = pLeftEyeRT->Release();
		}
	}
	pLeftEyeRT = 0;

	if (pRightEyeDS)
	{
		refCount = 1;
		while(refCount)
		{
			refCount = pRightEyeDS->Release();
		}
	}
	pRightEyeDS = 0;

	if (pLeftEyeDS)
	{
		refCount = 1;
		while(refCount)
		{
			refCount = pLeftEyeDS->Release();
		}
	}
	pLeftEyeDS = 0;

	if (pBackBufferSurface)
	{
		refCount = 1;
		while(refCount)
		{
			refCount = pBackBufferSurface->Release();
		}
	}
	pBackBufferSurface = 0;

	if (pBackBufferDepthStencil)
	{
		refCount = 1;
		while(refCount)
		{
			refCount = pBackBufferDepthStencil->Release();
		}
	}
	pBackBufferDepthStencil = 0;
}